/*
--------------Copyright (c) 2009-2015 H a r a l d  A c h i t z---------------
-----------< h a r a l d dot a c h i t z at g m a i l dot c o m >------------
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
-----------------------------------------------------------------------------
*/



#ifndef SQ3_SQLCOMMAND_HPP
#define SQ3_SQLCOMMAND_HPP

#include <string>
#include <sq3/dataset.hpp>
#include <sq3/dbvalue.hpp>
#include <sq3/queryhandler.hpp>


struct sqlite3;
struct sqlite3_stmt;

namespace sq3 {

class QueryHandler;

namespace internal
{
  class Connection;
}


/**
 * \brief A compiled SQL command
 *  
 *  This class holds a compiled SQL statement so that it can be reused.
 *  	
 */
class Command final
{

public:

  /**
   * \brief Parameters of a SQL command
   *
   *  This class holds the parmeters of a SQL command (if there are some).
   *
   */
  class Parameters final : public DbValues
  {
      friend class Command;

      using DbValues::DbValues ;

      Parameters(DbValues&& row)
           : DbValues(std::move(row))
           {

           }

    public:



      /**
       * \brief get name of a parameter
       *
       * if a parameter has a name it can be requested with this function.
       *
       * \throw ErrOutOfRange if the index is invalid.
       *
       * \param  parameter index
       * \return the name (which is maybe empty)
       */
      std::string getName( size_t idx ) const;



    private:

      std::vector<std::string> _names;

  };




  private:
    friend class Database;
    
    using ConnectionPtr = std::shared_ptr<internal::Connection>;

    Command(ConnectionPtr connection, 
                const std::string& sql, 
                DbValues parameters = {});

    




  public:

    Command(Command&& other) ;

    Command() = delete;
    Command(const Command&) = delete;
    Command operator=(const Command&) = delete;
    Command operator=(Command&&) = delete;


    /**
     * \brief d'tor.
     *  calls Release
     */
    ~Command(); // cause of unique pointer, check if protected is possible



    Dataset select();
    Dataset select(const DbValues& parameters);
    Dataset select(const Types& types);
    Dataset select(const DbValues& parameters, Types types);

    void execute( );
    void execute( const DbValues& parameters );
    void execute( QueryHandler& sqh ) ;
    void execute( QueryHandler& sqh, const DbValues& parameters ) ;
    void execute( QueryCallback sqc ) ;
    void execute( QueryCallback sqc,const DbValues& parameters ) ;



    //@{
    /**
     * \brief Parameters of command.
     *
     * \return command's parameters
     */
    Parameters& getParameters() ;
    const Parameters& getParameters() const ;
    //@}

    //@{
    /**
     * \brief get Parameter at given index.
     * \throw ErrOutOfRange if index is invalid
     * \return command's parameters
     */
    DbValue& getParameter(int idx);
    const DbValue& getParameter(int idx) const;
    //@}

    /**
     * \Set new parameter values
     * \throw ErrTypeMissMatch if size of values differs from the actual
     * parameter size or DbValue assignment throws.
     * \see DbValue
     *
     */
    void setParameters (const DbValues& values) ;

    /**
     * \Set new parameter values
     * \throw ErrTypeMissMatch if size of values differs from the actual
     * parameter size.
     *
     * In contrast to setParameters, where the DbValue types must match,
     * this can be used to set other Types.
     */
    void resetParameters (const DbValues& values) ;


  private:

    void runStepOnRowLoop(std::function<bool(sqlite3_stmt*)> cb);

    ConnectionPtr _connection;
    sqlite3_stmt* _stmt;
    Parameters    _parameters;

	
};






}



#endif
