/*
--------------Copyright (c) 2009-2015 H a r a l d  A c h i t z---------------
-----------< h a r a l d dot a c h i t z at g m a i l dot c o m >------------
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
-----------------------------------------------------------------------------
*/


#ifndef SQ3_DATABASE_HPP_
#define SQ3_DATABASE_HPP_


#include <string>
#include <memory>
#include <sq3/command.hpp>
#include <sq3/config.hpp>
#include <sq3/dataset.hpp>
#include <sq3/dbvalue.hpp>


struct sqlite3;

namespace sq3
{
  namespace internal
  {
    class Connection;
  }


  /**
   * \brief represents a SQLite3 Database
   *
   *
   * Encapsulate some/the most - useful methods for a SQLite3 database object.
   *
   *
   * \note for valgrind: http://www.sqlite.org/cvstrac/tktview?tn=3428 \n
   *  since I don't know where to put this note else I place it just here
   */
  class Database
  {

  public:

    /**
     * \brief c'tor
     *
     * Construction of this object opens, or creates a sqlite3 database.
     * The exact behavior can be fine tuned using the openFlags parameter.
     *
     * \param name database name.
     * \n If name is ":memory:" than the database will be an in memory database.
     * \n If name has a size of 0 than the database will be private on disk
     * database.
     * \n Otherwise name will be interpreted as a file.
     *
     * \param openFlags if no flags are given, the defaults are
     * SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE.
     *
     * \sa https://www.sqlite.org/c3ref/open.html
     *
     * \thorw SQLite3Error if the database can not be opened
     */
    Database( const std::string& name, int openFlags = 0 );

    /**
     * \brief d'tor.
     */
    virtual ~Database() noexcept ;

    Database(const Database&) = delete ;
    Database& operator=(const Database&) = delete ;
    Database& operator=( Database&&) = delete ;

    /**
     * \brief m'tor
     *
     * A Database is moveable
     */
    Database(Database&&) noexcept  ;

    /**
     * \brief create a SqlCommand instance.
     *
     * \return a SqlCommand instance
     */
    Command command(const std::string& sql,
                       const DbValues& = {});



    /**
     * \brief Execute one or more SQL statements.
     *
     * \n throws SQLite3Error in case of other problems.
     *
     * \param sql SQL Statements
     */
    void execute(const std::string& sql);

    /**
     * \brief Execute one or more SQL statements.
     *
     * Overload for const char to avoid a copy for larg command like
     * create database scripts which are often static  char*
     *
     * \n throws SQLite3Error in case of other problems.
     *
     * \param sql SQL Statements
     */
    void execute(const char* sql);


    /**
     * \brief Execute one SQL statements and apply a SqlQueryHandler.
     *
     * \n throws SQLite3Error in case of other problems.
     *
     * \param sql SQL Statements
     * \param SqlQueryHandler to handle query result
     */
    void execute(const std::string& sql,
                       QueryHandler& sqh);

    /**
     * \brief Execute one SQL statements and apply a QueryCallback.
     *
     * \n throws SQLite3Error in case of other problems.
     *
     * \param sql SQL Statements
     * \param QueryCallback to handle query result
     */
    void execute(const std::string& sql,
                       QueryCallback sqc);


    /**
     * \brief Execute a SQL query and returns the result .
     *
     * \n throws SQLite3Error in case of other problems.
     *
     * \param sql SQL Statements
     * \return a Dataset with the result.
     */
    Dataset select(const std::string& sql);

    /**
     * \brief Execute a SQL query and returns the result .
     *
     * \n throws SQLite3Error in case of other problems.
     *
     * \param sql SQL Statements
     * \param types wanted types of the returned Dataset
     * \return a Dataset with the result.
     */
    Dataset select(const std::string& sql, const Types& types);

    /**
     * \brief select a single value form the database.
     *
     * This is a quick way to get single value results for queries like
     *  "SELECT COUNT(*) FROM someTable;" \n
     *
     * This function returns the first column of the first result row of the
     *  given query.
     * If query returns no rows, the return value will be null.
     *
     * \param sql SQL statement
     * \return DbValue of the first available result.
     */
    DbValue selectValue(const std::string& sql);


    /**
     * \brief select a single typed value form the database.
     *
     * This is a quick way to get single value results for queries like
     *  "SELECT COUNT(*) FROM someTable;" \n
     *
     * This function returns the first column of the first result row of the
     *  given query.
     * If query returns no rows, the return value will be null.
     *
     * \throw ErrTypeMisMatch if return type differs from requested type
     *
     * \param sql SQL statement
     * \param type resutl has to be of that Type
     *
     * \return DbValue of the first available result.
     */
    DbValue selectValue(const std::string& sql, Type type);

    /**
     * \brief Returns last SQLite3 extended result code
     *
     *  \note SQLite3 result code in case of a failure is triggered within a
     *   SQLite3Error.
     *
     *  The return value is only valid if the last call to the failed.
     *
     * \note This function returns extended result codes
     *
     * \return SQLite3 extended error code
     */
    int getMostRecentErrCode();


    /**
     * \brief Returns most recent error message.
     *
     *
s     * \return SQLite3 error message.
     */
    std::string getMostRecentErrMsg();


    /**
      * \brief Wraps sqlite3_errstr() function
      *
      * The sqlite3_errstr() interface returns the English-language text
      * that describes the result code, as UTF-
      *
      * \return Count of changed rows
      */
    std::string getErrStr(int reslutcode);

    /**
     * \brief Returns number of row that have changed since database opening.
     *
     * Returns the number of rows that have been changed
     * Through successful SQL INSERT/UPDATE or DELETE statements since database
     *  was opened.
     *
     * \return Count of changed rows
     */
    std::size_t getTotalChanges();

    /**
     * \brief Rows that have been changed from the most recent SQL statement.
     *
     * Returns the number of rows that have been changed from the most recent
     * successful SQL INSERT/UPDATE or DELETE statement.
     *
     * \return Count of changed rows
     */
    std::size_t getRecentlyChanged();

    /**
     * \brief returns rowid  of the most recent successful INSERT statement
     *
     * Returns the rowid (ROWID, OID, or _ROWID_ or the column of
     * type INTEGER PRIMARY KEY )
     * of the most recent successful INSERT into the database.
     * \n If no successful INSERTs have ever occurred on that database,
     * zero is returned.
     *
     * \return rowid
     */
    int64_t getLastInsertRowid();

    /**
     * \brief returns value of SQLITE_THREADSAFE compilation option
     *
     *  see http://www.sqlite.org/compile.html about additional informations
     *
     * \return 0 or 1 or 2
     */
    int getThreadSafeCompileOption();

    /**
     * \brief Transaction Guard
     *
     * Scope guard for transaction.
     * If an instance of this class goes out of scope and commit has
     * not been called, it will call Rollback.
     */
    class TransactionGuard
    {
      sq3::Database& _db;
      bool _commit;

      TransactionGuard(Database& );
      friend class Database;

    public:
      TransactionGuard(TransactionGuard&&) ;
      ~TransactionGuard();
      void commit();

      TransactionGuard(const TransactionGuard&) = delete;
      TransactionGuard& operator=(const TransactionGuard&) = delete;
      TransactionGuard& operator=(TransactionGuard&&) = delete;


    };

    /**
     * \brief Create a TransactionGuard
     */
    TransactionGuard beginTransaction();


  protected:
    using ConnectionPtr = std::shared_ptr<internal::Connection>;
    ConnectionPtr _connection;

  };


  namespace internal
  {
    /**
     * \brief Holds the sqlite3 connection pointer.
     *
     * This is the link between Database and Command , and only used internal
     * \internal
     */
    class Connection
    {
      friend class sq3::Database;

    public:

     /// return db pointer
     sqlite3* db() ;

     /// return if connection is open
     bool isValid();

     ///  throw ErrNoConnection if not valid
     void ensureValid();


    private:
      Connection(sqlite3* p);
      Connection(Connection&&) = default;

      Connection(const Connection&) = delete;
      Connection& operator=(const Connection&) = delete;
      Connection& operator=(Connection&&) = delete;


      sqlite3* sq3ptr ;

    };
  }


}

#endif /* ...DATABASE_HPP_ */
