/*
 --------------Copyright (c) 2012-2015 H a r a l d  A c h i t z---------------
 -----------< h a r a l d dot a c h i t z at g m a i l dot c o m >------------
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 -----------------------------------------------------------------------------
 */

#ifndef SQ3_DATASET_HPP_
#define SQ3_DATASET_HPP_

#include <vector>
#include <map>

#include <sq3/dbvalue.hpp>
#include <sq3/error.hpp>
#include <sq3/queryhandler.hpp>

namespace sq3{



/**
 * \brief A utility for processing the result queries.
 *
 * This class is a RowHandler that loads the result of a Query into a list
 *  of DbValues.
 * The loaded list can be browsed and the loaded values can be accessed.
 *
 * A Dataset can either be created:
 *  - giving a DbValuesTypeList that describes the fields \n
 *   Giving a DbValuesTypeList the types of DbValue will use the given storage
 *   type when reading
 *   the data from sqlite. \n
 *   When the Dataset becomes populated with data the field count will be
 *   validated.
 *   In case the number of fields is different sq3::ErrTypeMisMatch will
 *   be thrown.
 *
 *  - Without any specification \n
 *  In this case all fields will be DsVariantField ,
 *  using the storage type sqlite reports for the actual value.
 *
 *
 *
 */
class Dataset final: public Container<std::vector<DbValues>>
                   , public QueryHandler
{


  friend class Database;

public:

  /**
   * \brief c'tor
   * All fields will be DsVariantField , using the storage type sqlite reports for the actual value.
   * Field count will be detected and applied.
   */
  Dataset() noexcept;
  /**
   * \brief c'tor giving a DbValuesTypeList as description
   * Types of DbValue will use the given description when creating the DbValue list.
   * If the given list is not empty, field count will be validated when the actual instance becomes populated with data.
   *
   */
  Dataset(Types fl);


  Dataset(const Dataset& other);
  Dataset(Dataset&& other) noexcept (
         std::is_nothrow_move_constructible<Container<DbValues>>::value &&
         std::is_nothrow_move_constructible<Types>::value &&
         std::is_nothrow_move_constructible<std::vector<std::string>>::value);
  // is noexcept, but I leave this for now ..

  Dataset& operator=(const Dataset& other);
  Dataset& operator=(Dataset&& other);




  /**
   * \brief Clear all states.
   * Removes loaded data so that the actual instance can be reused/refilled.
   *
   */
  void reset();
  
  /**
   * \brief Clear all states.
   * Removes loaded data and sets a new specification for the field description
   * so that the actual instance
   * can be reused for populate with a different select statement / sql command.
   * Passing an empty DbValuesTypeList mean that all fields will be
   * DsVariandField and field count will be detected.
   *
   */
  void reset(const Types& fl);


  /**
   * \brief Merge an other Dataset.
   *
   * Appends the data of the given Dataset to the end of the actual data.
   * The field names and types of the given Dataset must match the actuals one.
   *
   * \throw sq3::ErrTypeMisMatch if field names types are not equal or
   * size differs.
   *
   * \param other Dataset which shall be added to this one.
   */
  void merge(const Dataset& other) ;


  /**
   * \brief Merge DbValues.
   *
   * Appends the DbValues to the end of the actual data.
   *
   * \throw sq3::ErrTypeMisMatch if size differs from existing row size.
   *
   * \param  DbValues which shall be added.
   */
  void merge(const DbValues& other) ;

  /**
   * \brief Get the index of a field by namespace
   *
   * \throw OutOfRanage if name is not found
   * \param name field name
   * \return field index
   */
  std::size_t getIndex(const std::string& name) const;


protected:
  bool onRow(const QueryRow& queryrow) override;
  void onStart() override;
  void onEnd() override;
  

  Types _fieldtypes;
  std::vector<std::string>_names;

  
};

}

#endif 
