/*
 --------------Copyright (c) 2009-2015 H a r a l d  A c h i t z---------------
 -----------< h a r a l d dot a c h i t z at g m a i l dot c o m >------------
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 -----------------------------------------------------------------------------
 */

#ifndef SQ3_DbValue_HPP_
#define SQ3_DbValue_HPP_


#include <memory>
#include <string>
#include <vector>
#include <sq3/error.hpp>
#include <sq3/utils/container.hpp>


namespace sq3 {

/**
 * Enumeration of different value types
 *
 * These are the types known by sqlite.
 * They are used when reading from or writing to a database.
 *
 */
enum class Type
{
  Variant = 0, //!< takes any type by value,
  Int = 1, //!< Int value
  Real = 2, //!< Real value
  Text = 3, //!< Text value
  Blob = 4, //!< Blob vale
  Null = 5    //!< Null

};


class Types : public Container<std::vector<Type>>
{

public:

  using conatiner_type = typename Container::conatiner_type  ;
  using iterator = typename conatiner_type::iterator;
  using const_iterator = typename conatiner_type::const_iterator;
  using value_type = typename conatiner_type::value_type;
  using reference = typename conatiner_type::reference;
  using const_reference = typename conatiner_type::const_reference;
  using size_type = typename conatiner_type::size_type;

  using Container::Container ;


  void swap (Types& other)
  {
     using std::swap;
     swap (_cont, other._cont);
  }

};


/**
 * A type for binary data
 */
using Blob = std::vector<char>;


/**
 *  \brief
 *
 *  A DbValues has a type and a value and guarantees that
 *  the value fits to the type.
 *
 *  Using this type for communication with the database can ensure either
 *  type guarantee or give the flexibility of variant types.
 *
 */
class DbValue
{

public:

  /** Constructor
   *
   *  Constructs a type  and the value is null.
   *  if   Type::Null is given, the type will be a variant.
   */
  DbValue(Type type) noexcept;

  //@{
  /** Constructor
   *
   *  This constructor takes a initialization value, and optionally a type
   *  which can only be Type::Variant or the default one.
   *
   *  \throw ErrTypeMisMatch if given type is incompatible
   */
  DbValue(int val, Type type = Type::Int);
  DbValue(int64_t val, Type type = Type::Int);
  DbValue(std::string val, Type type = Type::Text);
  DbValue(double val, Type type = Type::Real);
  DbValue(Blob val, Type type = Type::Blob) ;
  //@}

  ~DbValue() noexcept;

  DbValue(const DbValue& other) noexcept;
  DbValue(DbValue&& other) noexcept;

  //@{
  /** Assignment
   *  \throw ErrTypeMisMatch if getType is incompatible
   *  \note , only value assignment happens here,
   *  the type does not change, storage type might change in case of type
   *  is a variant.
   *
   */
  DbValue& operator=(const DbValue& other);
  DbValue& operator=(DbValue&& other) ;
  DbValue& operator=(int val);
  DbValue& operator=(const int64_t& val);
  DbValue& operator=(const double& val);
  DbValue& operator=(const std::string& val);
  DbValue& operator=(const Blob& val);
  void set(int val);
  void set(int64_t val);
  void set(const std::string& val);
  void set(double val);
  void set(const Blob& val) ;

  //@}


  //@{
  /** value access
   *  \throw ErrTypeMisMatch if getType is incorrect
   */
  const int64_t& getInt() const;
  const double& getReal() const;
  const std::string& getText() const;
  const Blob& getBlob() const;
  //@}

  //@{
  /** compare value
   *
   */
  bool operator==(const int val) const  ;
  bool operator==(const int64_t& val) const  ;
  bool operator==(const std::string& val) const  ;
  bool operator==(const double& val) const  ;
  bool operator==(const Blob& val) const  ;

  bool operator!=(const int val) const ; 
  bool operator!=(const int64_t& val) const ;
  bool operator!=(const std::string& val) const ;
  bool operator!=(const double& val) const  ;
  bool operator!=(const Blob& val) const  ;

  //@}

  //@{
  /** moves the current value into the return value
   *  To move large string or Blob object.
   *
   *  After calling this function the value will be Null.
   */
  std::string ejectText();
  Blob        ejectBlob();
  //@}

  ///set DbValue value to NULL
  void setNull() ;
  /// returns if value is Null
  bool isNull() const { return _storageType == Type::Null ; }

  /** returns the Type
   *
   * This is the type with which the Value has been created
   * and will not change.
   */
  Type getType() const { return _type; }

  /** returns the underlying type
   *
   * If getType() is a Variant this property
   * returns the actual type information, otherwise it will be the same.
   */
  Type getStorageType() const { return _storageType; };


protected:
  
  const Type _type; 
  // type never changes, even == makes just the value, not the type,
 
  Type _storageType; 
  // that's the type for the union and what is applied in the db
  union {
    int64_t     _intstore; 
    double      _realstore;
    std::string _textstore;
    Blob        _blobstore;
  } ;

  void clearValue();

};


std::ostream&
operator<<(std::ostream& stm, const sq3::DbValue& v);



/**
 * \brief Fixed size DbValues
 */
class DbValues : public Container<std::vector<DbValue>>
{

public:

  using conatiner_type =  typename Container::conatiner_type  ;
  using iterator = typename conatiner_type::iterator;
  using const_iterator = typename conatiner_type::const_iterator;
  using value_type = typename conatiner_type::value_type;
  using reference = typename conatiner_type::reference;
  using const_reference = typename conatiner_type::const_reference;
  using size_type = typename conatiner_type::size_type;

  using Container::Container ;

  DbValues(const DbValues& row):Container(row){}

  DbValues(DbValues&& row) :Container(std::move(row)){}


  //@{
  /** assigment
   *  \throw ErrTypeMisMatch if size() is different
   *  \see assignment of DbValue
   */
  DbValues& operator=(const DbValues& row);

  DbValues& operator=(DbValues&& row) ;
  //@}


  void swap(DbValues& other)
  {
    ASSERT_EXCEPT (other.size () == size (), ErrTypeMisMatch) ;

    using std::swap;
    swap(_cont, other._cont);
  }


};






}

#endif /* DbValue_HPP_ */
