/*
--------------Copyright (c) 2009-2015 H a r a l d  A c h i t z---------------
-----------< h a r a l d dot a c h i t z at g m a i l dot c o m >------------
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
-----------------------------------------------------------------------------
*/

#ifndef SQ3__ERROR_HPP_
#define SQ3__ERROR_HPP_


#include <iosfwd>
#include <string>



namespace sq3{


  /// map error reason to enum value
  enum class ErrCode
  {
    SQL3Error = 1 , ///< sqlite3 error happened
    NO_CONNECTION = 2 , ///< no database
    OUT_OF_RANGE = 5 ,///< index op out of range
    TYPE_MISMATCH = 6 ,///< type cast problem
    NullValueAccess = 7 , ///< accessing a value that is Null
    UNEXPECTED = 99 ///< for everything that happens unexpected
  };
  


  class Error
  {

  public:
    Error( std::string info ) ;

#ifdef _MSC_VER 
    virtual ~Error() = default ;
#else
    virtual ~Error() noexcept = default ;
#endif

    virtual ErrCode id() const = 0 ;

    const std::string& info() const ;

  protected:
    const std::string _info;

    friend std::ostream& operator<<(std::ostream& os, const Error& e);

    virtual void toStream(std::ostream& os) const;
  };

  std::ostream& operator<< (std::ostream& os, const Error& e);


  template<ErrCode ec>
  class ErrType: public Error
  {

  public:
    ErrType( std::string msg = "" )
      : Error( std::move(msg) )
    {
    }

#ifdef _MSC_VER
    virtual ~ErrType() = default ;
#else
    virtual ~ErrType() noexcept = default ;
#endif


    ErrCode id() const override  { return ec ; }

  };
  
  /// error that will be thrown if no open database was available but needed
  using ErrNoConnection = ErrType<ErrCode::NO_CONNECTION> ;

  ///thrown if an index op is out of range
  using ErrOutOfRange = ErrType<ErrCode::OUT_OF_RANGE> ;

  ///thrown in case of an incorrect type usage
  using ErrTypeMisMatch = ErrType<ErrCode::TYPE_MISMATCH>;

  ///thrown in case of accessing a Null value field/parameter
  using ErrNullValueAccess = ErrType<ErrCode::NullValueAccess> ;

  ///thrown if something unexpected happened, mostly used by test tools and in debug mode
  using ErrUnexpected =  ErrType<ErrCode::UNEXPECTED>;

  
  /**
   * \brief packaging an error from sqlite3. 
   * This exception will be thrown if sqlite3 reports an error.
   * 
   * Holds the sqlite3 error code and the error message if it is available.
   * \see Database::getMostRecentErrMsg
   * \see Database::getMostRecentErrCode
   * 
   * \n Extended error codes can be obtained through Database class.
   * \see Database::getMostRecentExtendedErrCode 
   * \see Database::enableExtendedResultCodes 
   * 
   */
  class SQLite3Error : public ErrType<ErrCode::SQL3Error>
  {

      int m_sqlite_ec;
      std::string m_sqlite_msg;
      
  public: 
      /// c'tor
      SQLite3Error( int sq3ec, const char* sq3msg )
      : ErrType<ErrCode::SQL3Error>( std::string("") )
      , m_sqlite_ec( sq3ec ) 
      , m_sqlite_msg ( sq3msg )
      {
      }
      
      ///c'tor
      SQLite3Error( int sq3ec, const char* sq3msg , const std::string& msg)
      : ErrType<ErrCode::SQL3Error>( msg )
      , m_sqlite_ec( sq3ec ) 
      , m_sqlite_msg ( sq3msg )
      {
      }
      
#ifdef _MSC_VER
    virtual ~SQLite3Error() = default ;
#else
    virtual ~SQLite3Error() noexcept = default ;
#endif

      
      /// returns the sqlite3 error code
      int SQLiteErrorCode() const { return m_sqlite_ec ; }
      
      /// returns the sqlite3 error message (if it was aviable)
      const std::string& SQLiteErrorMessage() const { return m_sqlite_msg ; }
  
  private:
      virtual void toStream(std::ostream& os)  const override ;
      
  };

#define  ASSERT_EXCEPT( exp , except ) if ( !( exp ) ) \
  throw  except( std::string(__PRETTY_FUNCTION__) + ": " + #exp )


} // ns 


#endif /* ...ERROR_HPP_ */
