/*
--------------Copyright (c) 2009-2015 H a r a l d  A c h i t z---------------
-----------< h a r a l d dot a c h i t z at g m a i l dot c o m >------------
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
-----------------------------------------------------------------------------
*/


#ifndef SQ3_SQLQUERYHANDLER_HPP
#define SQ3_SQLQUERYHANDLER_HPP

#include <sq3/dbvalue.hpp>

struct sqlite3_stmt;

namespace sq3 {

class QueryRow
{

public:

  ///c'tor
  QueryRow( sqlite3_stmt* stmt  ) : _stmt(stmt) {}

  ///d'tor
  ~QueryRow() = default ;

  ///c'tor
  QueryRow(const QueryRow&) = default ;

  ///c'tor
  QueryRow(QueryRow&&) = default ;

  // they should not be needed, even if they would not harm ..
  QueryRow& operator=(const QueryRow&) = delete ;
  QueryRow& operator=( QueryRow&&) = delete ;



  /**
   * \brief returns column name of given index
   *
   *  \return column name or an empty string if index is invalid
   */
  std::string getColumnName( int idx ) const;

  /// returns number of columns in the statement.
  int getColumnCount() const;

  /// returns getColumnCount() as size_t
  size_t size() const { return static_cast<size_t>(getColumnCount()) ; }


  /**
   * \brief access ColIdx object
   * \param idx wanted index
    *  The return DbValue will always be a Type::Variant and
   *  DbValue::getStorageType() will be set to the sqlite3 storage type.
   * \throw sq3::ErrOutOfRange if idx is to high,
   * \return DbValue of requested column
   */
  DbValue at(int idx) const ;


  /**
   * \brief unchecked index access
   *
   *  Will behave undefined if index is invalid.
   *  The return DbValue will always be a Type::Variant and
   *  DbValue::getStorageType() will be set to the sqlite3 storage type.
   * \param idx wanted index
   * \return DbValue of requested column
   */
  DbValue operator[](const int idx) const;

  /// return column names
  std::vector<std::string> getNames() const;


  /// return all columns at once
  DbValues getRow() const;


  /// return all columns at once
  DbValues getRow(const Types& types) const;


  /**
   * \brief get underlying sqlite3_stmt
   *
   * user defined QueryCallbacks might have their own way to do things,
   * so this getter provides access to the underlying sqlite3_stmt.
   *
   * \return underlying sqlite3_stmt
   */
  sqlite3_stmt* get_stmt() const;

private:
  sqlite3_stmt* _stmt ;


};




  /**
   * \brief function object for handling a query result.
   *
   * Lightweight form of a RowHandler taking only Columns as argument.
   * \see SqlQueryHandler::onRow
   *
   */
  using QueryCallback =  std::function<bool(QueryRow& )> ;


/**
 * \brief Callback for SQL Select statements
 * 
 * This interface can be used to process through the result of a SQL statement.
 *  
 * 
 */
class QueryHandler
{

    friend class Command;

  protected:

    /**
     * \brief d'tor
     */
    virtual ~QueryHandler() noexcept = default ;

    /**
     * \brief handles one row of the result from  a SELECT statement
     *
     *
     * @param cols RowColumns object
     * @param stmt sqlite3_stmt to be able to use native sqlite calls
     * @return false if processing the query result shall stop
     *       \n true otherwise
     */
    virtual bool onRow(const QueryRow& queryrow) = 0;

    /**
     * \brief called before a query result will be processed
     */
    virtual void onStart(){}

    /**
    * \brief called after a query result has been processed
    */
    virtual void onEnd(){}


  private:

    /**
     * \brief Internal call, prepare OnHandleRow call
     *
     *
     * \n raise rowcout + 1
     * \n create the RowColumns object
     * \n calls OnHandleRow
     *
     * @param stmt actual sqlite3_stmt
     * @return result of OnHandleRow implementation
     */
    bool handleRow( sqlite3_stmt* stmt ) ;


};






}



#endif
