/*
 --------------Copyright (c) 2009-2015 H a r a l d  A c h i t z---------------
 -----------< h a r a l d dot a c h i t z at g m a i l dot c o m >------------
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 -----------------------------------------------------------------------------
 */

#ifndef SQ3_CONTAINER_HPP_
#define SQ3_CONTAINER_HPP_


#include <sq3/error.hpp>


namespace sq3
{

/**
 * \brief Wrapper to provide begin, end and random access of a container
 *
 * A Container has begin, end and random access.
 *
 * This class makes these methods accessible and hides the others, like a
 * push_back, erase,...
 * A derived class can decide which other methods shall become visible, or not.
 *
 *
 *
 */
template<typename ContainerType>
class Container
{
 public:

  Container() noexcept{}
     

  Container(ContainerType c)
   noexcept(std::is_nothrow_move_constructible<ContainerType>::value)
   :_cont(std::move(c))
  {
  }

  Container(std::initializer_list<typename ContainerType::value_type> l)
   :_cont(std::move(l))
  {
  }  


  Container(const Container&) = default;
  Container& operator=(const Container&) = default;

  Container (Container&&) 
    noexcept(std::is_nothrow_move_constructible<ContainerType>::value)=default;

  Container& operator=(Container&&) = default;

  ///d'tor
  virtual ~Container()
     noexcept(std::is_nothrow_destructible<ContainerType>::value) = default;

  //@{
  /// usual typedefs
  using conatiner_type =  ContainerType ;
  using iterator = typename conatiner_type::iterator;
  using const_iterator = typename conatiner_type::const_iterator;
  using value_type = typename conatiner_type::value_type;
  using reference = typename conatiner_type::reference;
  using const_reference = typename conatiner_type::const_reference;
  using size_type = typename conatiner_type::size_type;
  //@}

  /// additional typedef
  using container_type = ContainerType;

  //@{
  /// iterator access
  iterator begin() { return std::begin(_cont);}

  const_iterator begin() const { return _cont.cbegin();}

  iterator end() { return std::end(_cont);}

  const_iterator end() const { return _cont.cend();}

  const_iterator cbegin() const { return _cont.cbegin();}

  const_iterator cend() const { return _cont.end();};
  //@}

  /// returns number of elements
  size_type size() const { return _cont.size();}
  
 
  //@{
  /** checked random access
   * \throw ErrOutOfRange if indes is invalid
   */
  reference at(size_t i){
    try{
        return _cont.at(i);
    } catch(...) {
        throw ErrOutOfRange("no data at: " + std::to_string(i)) ;
    }
  }
  const_reference at(size_t i) const {
    try{
        return _cont.at(i);
    } catch(...) {
        throw ErrOutOfRange("no data at: " + std::to_string(i)) ;
    }
  }
  //@}

  //@{
  /// unchecked random access
  reference operator[](size_t i) {  return _cont[i] ;}

  const_reference operator[](size_t i) const {  return _cont[i] ;}
  //@}


 protected:
  ContainerType _cont;

};



}

#endif /* SQ3_CONTAINER_HPP_ */
