/*
 --------------Copyright (c) 2009-2015 H a r a l d  A c h i t z---------------
 -----------< h a r a l d dot a c h i t z at g m a i l dot c o m >------------
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 -----------------------------------------------------------------------------
 */

#include <sq3/database.hpp> // connection
#include <sq3/command.hpp>
#include <sq3/error.hpp>
#include <sqlite3.h>

#include <functional>

namespace sq3
{

  namespace
  {

    void
    runStmtLoop (sqlite3_stmt* stmt, std::function<bool(sqlite3_stmt*)> cb)
    {

      bool loop = true;
      while (loop)
        {

          int rc = sqlite3_step (stmt);

          switch (rc)
            {

            case SQLITE_OK :
            case SQLITE_DONE :
              sqlite3_reset (stmt);
              loop = false;
              break;

            case SQLITE_ROW :
              loop = cb (stmt);
              break;

            default :
              auto db = sqlite3_db_handle (stmt);
              SQLite3Error sq3error (rc, sqlite3_errmsg (db));
              sqlite3_reset (stmt);
              throw sq3error;

            }

        }

    }
//------------------------------------------------------------------------------

    void
    bind (sqlite3_stmt* stmt, Command::Parameters& parameters)
    {
      int curParaNr = 0;
      for (DbValue& val : parameters)
        {
          curParaNr += 1; // sqlite starts at 1

          try
            {
              int rc;

              switch (val.getStorageType ())
                {
                case Type::Int :
                  rc = sqlite3_bind_int64 (stmt, curParaNr, val.getInt ());
                  break;

                case Type::Real :
                  rc = sqlite3_bind_double (stmt, curParaNr, val.getReal ());
                  break;

                case Type::Text :
                  // note, i do not want \0 in the db so take size
                  // SQLITE_TRANSIENT would copy the string , is unwanted here
                  rc = sqlite3_bind_text (stmt, curParaNr,
                                          val.getText ().c_str (),
                                          val.getText ().size (),
                                          SQLITE_STATIC);

                  break;

                case Type::Blob :
                  rc = sqlite3_bind_blob (stmt, curParaNr,
                                          &(val.getBlob () [0]),
                                          val.getBlob ().size (),
                                          SQLITE_STATIC);

                  break;

                case Type::Null :
                  rc = sqlite3_bind_null (stmt, curParaNr);
                  break;

                default :
                  throw ErrUnexpected ();
                }

              if (rc != SQLITE_OK)
                throw sq3::SQLite3Error (rc, "");
            }
          catch (...)
            {
              throw;
            }

        }

    }

  } // ns
//------------------------------------------------------------------------------

  auto
  Command::Parameters::getName (size_t idx) const -> std::string
  {

    if (idx >= _names.size ())
      throw ErrOutOfRange ("idx " + std::to_string (idx));

    return _names [idx];
  }

//------------------------------------------------------------------------------

  Command::Command (ConnectionPtr connection,
                    const std::string& sql,
                    DbValues parameters) :
      _connection (std::move (connection)),
      _stmt (nullptr),
      _parameters (std::move (parameters))
  {

    const char* unussedSQL = 0;
    // unused stays unused,

    int rc = sqlite3_prepare_v2 (_connection->db (), sql.c_str (), -1, &_stmt,
                                 &unussedSQL);

    if (rc != SQLITE_OK)
      {
        SQLite3Error sq3error (rc, sqlite3_errmsg (_connection->db ()));
        _stmt = nullptr;
        throw sq3error;
      }

    const size_t paracount = sqlite3_bind_parameter_count (_stmt);

    if (paracount > 0)
      {
        if (_parameters.size () != 0)
          {
            if (paracount != _parameters.size ())
              throw ErrTypeMisMatch ("Incorrect parameter count");
          }
        else
          {
            _parameters._cont.resize (paracount, Type::Variant);
          }

        _parameters._names.resize (paracount);
        for (size_t i = 0; i < paracount; ++i)
          {
            const char* chrname = sqlite3_bind_parameter_name (_stmt, i + 1);
            _parameters._names [i] = chrname ? chrname : "";
          }
      }

  }
//------------------------------------------------------------------------------

  Command::Command (Command&& other) :
      _connection (std::move (other._connection)),
      _stmt (other._stmt),
      _parameters (std::move (other._parameters))
  { // clear stm so that d'tor ot other does no action
    other._stmt = nullptr;
  }

//------------------------------------------------------------------------------
  Command::~Command ()
  {

    if (_stmt) // its not a moved from zombi
      {
        if (_connection->isValid ()) // otherwise database will have done this
          {
            sqlite3_finalize (_stmt);
          }
      }

  }
//------------------------------------------------------------------------------

  Dataset
  Command::select ()
  {
    Dataset ds;
    execute (ds);
    return ds;
  }
//------------------------------------------------------------------------------

  Dataset
  Command::select (const DbValues& parameters)
  {
    Dataset ds;
    execute (ds, parameters);
    return ds;
  }
//------------------------------------------------------------------------------

  Dataset
  Command::select (const Types& types)
  {
    Dataset ds { std::move (types) };
    execute (ds);
    return ds;
  }
//------------------------------------------------------------------------------

  Dataset
  Command::select (const DbValues& parameters, Types types)
  {
    Dataset ds { std::move (types) };
    execute (ds, parameters);
    return ds;
  }
//------------------------------------------------------------------------------

  void
  Command::execute ()
  {
    _connection->ensureValid ();

    if (_parameters.size () > 0)
      bind (_stmt, _parameters);

    auto donothing = [](sqlite3_stmt*)->bool
      { return false;};

    runStmtLoop (_stmt, donothing);
  }
//------------------------------------------------------------------------------

  void
  Command::execute (const DbValues& parameters)
  {
    _connection->ensureValid ();

    setParameters (parameters);

    bind (_stmt, _parameters);

    auto donothing = [](sqlite3_stmt*)->bool
      { return false;};

    runStmtLoop (_stmt, donothing);
  }
//------------------------------------------------------------------------------

  void
  Command::execute (QueryHandler& sqh)
  {
    _connection->ensureValid ();

    if (_parameters.size () > 0)
      bind (_stmt, _parameters);

    sqh.onStart ();

    auto applyrh = [&](sqlite3_stmt* stm)->bool
      {
        return sqh.handleRow(stm);
      };

    runStmtLoop (_stmt, applyrh);

    sqh.onEnd ();

  }
//------------------------------------------------------------------------------
  void
  Command::execute (QueryHandler& sqh, const DbValues& parameters)
  {
    _connection->ensureValid ();

    setParameters (parameters);

    bind (_stmt, _parameters);

    sqh.onStart ();

    auto applyrh = [&](sqlite3_stmt* stm)->bool
      {
        return sqh.handleRow(stm);
      };

    runStmtLoop (_stmt, applyrh);

    sqh.onEnd ();
  }
//------------------------------------------------------------------------------
  void
  Command::execute (QueryCallback sqc)
  {
    _connection->ensureValid ();

    if (_parameters.size () > 0)
      bind (_stmt, _parameters);

    auto applyrcb = [&](sqlite3_stmt* stm)->bool
      {
        QueryRow queryrow( stm );
        return sqc( queryrow );
      };

    runStmtLoop (_stmt, applyrcb);

  }
//------------------------------------------------------------------------------
  void
  Command::execute (QueryCallback sqc, const DbValues& parameters)
  {
    _connection->ensureValid ();

    setParameters (parameters);

    bind (_stmt, _parameters);

    auto applyrcb = [&](sqlite3_stmt* stm)->bool
      {
        QueryRow queryrow( stm );
        return sqc( queryrow );
      };

    runStmtLoop (_stmt, applyrcb);

  }
//------------------------------------------------------------------------------

  auto
  Command::getParameters () -> Parameters&
  {
    return _parameters;
  }
//------------------------------------------------------------------------------

  auto
  Command::getParameters () const -> const Parameters&
  {
    return _parameters;
  }
//------------------------------------------------------------------------------

  DbValue&
  Command::getParameter (int idx)
  {
    return _parameters.at (idx);
  }
//------------------------------------------------------------------------------

  const DbValue&
  Command::getParameter (int idx) const
  {
    return _parameters.at (idx);
  }
//------------------------------------------------------------------------------

  void
  Command::setParameters (const DbValues& values)
  {
    if (values.size () != _parameters.size ())
      {
        throw ErrTypeMisMatch ("parameter size incorrect");
      }

    for (size_t i = 0; i < values.size (); ++i)
      {
        _parameters [i] = values [i];
      }

  }
//------------------------------------------------------------------------------

  void
  Command::resetParameters (const DbValues& values)
  {
    auto tmp = values;
    _parameters.swap (tmp);

  }

} // ns

