/*
 --------------Copyright (c) 2009-2015 H a r a l d  A c h i t z---------------
 -----------< h a r a l d dot a c h i t z at g m a i l dot c o m >------------
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 -----------------------------------------------------------------------------
 */

#include <sq3/database.hpp>
#include <sq3/error.hpp>
#include <sqlite3.h>

namespace sq3
{

//----------------------------------------------------------------------------
// implement functions from config.h
  const char*
  sqliteCompiledVersion ()
  {
    return SQLITE_VERSION;
  }
  int
  sqliteCompiledVersionNumber ()
  {
    return SQLITE_VERSION_NUMBER;
  }

  const char*
  sqliteRuntimeVersion ()
  {
    return sqlite3_libversion ();
  }

  int
  sqliteRuntimeVersionNumber ()
  {
    return sqlite3_libversion_number ();
  }
//----------------------------------------------------------------------------

  namespace
  { // internal helper

    void
    close_sqlite3 (sqlite3* db) noexcept
    {
      if (db == nullptr)
        return;

      // total clean up to be sure nothing left.
      auto stm = sqlite3_next_stmt (db, 0);
      while (stm != nullptr)
        {
          sqlite3_finalize (stm);
          stm = sqlite3_next_stmt (db, 0);
        }

      // if  busy, use v2 for garbed collecting,
      if (sqlite3_close (db) != SQLITE_OK)
        { // but that should 'never' happen :-)
          sqlite3_close_v2 (db);
        }

    }
//-----------------------------------------------------------------------------

  }//ns anno
//------------------------------------------------------------------------------

  namespace internal
  {

    Connection::Connection (sqlite3* p) :
        sq3ptr (p)
    {
    }

    sqlite3*
    Connection::db ()
    {
      return sq3ptr;
    }

    bool
    Connection::isValid ()
    {
      return sq3ptr != nullptr;
    }

    void
    Connection::ensureValid ()
    {
      if (sq3ptr == nullptr)
        throw ErrNoConnection ();
    }

  } // ns internal

  Database::Database (const std::string& name, int openFlags) :
      _connection ()
  {

    if (openFlags == 0)
      openFlags = SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE;

    sqlite3* db = nullptr;
    auto sq3rc = sqlite3_open_v2 (name.c_str (), &db, openFlags, 0);

    if (sq3rc != SQLITE_OK)
      {
        SQLite3Error sq3error (sq3rc, sqlite3_errmsg (db));
        sqlite3_close (db);
        throw sq3error;
      }

    _connection.reset (new internal::Connection (db));

    sqlite3_extended_result_codes (_connection->sq3ptr, true);

  }
//------------------------------------------------------------------------------
  Database::Database (Database&& other) noexcept
  : _connection(std::move(other._connection))
    {
      other._connection.reset();
    }

//------------------------------------------------------------------------------
  Database::~Database () noexcept
  {
    if (_connection != nullptr)
      {
        close_sqlite3 (_connection->sq3ptr);

        // if there are somewhere commands arround, they need to know ..
        _connection->sq3ptr = nullptr;
      }
  }

//------------------------------------------------------------------------------

  Command
  Database::command (const std::string& sql, const DbValues& parameters)
  {
    return
      { _connection, sql, parameters};
  }
//------------------------------------------------------------------------------

  void
  Database::execute (const std::string& sql)
  {
    execute (sql.c_str ());
  }
//------------------------------------------------------------------------------

  void
  Database::execute (const char* sql)
  {
    char* dbMsg = 0;

    int rc = sqlite3_exec (_connection->sq3ptr, sql, 0, 0, &dbMsg);

    if (rc != SQLITE_OK)
      {
        SQLite3Error sq3error (rc, dbMsg);

        sqlite3_free (dbMsg);

        throw sq3error;
      }

    sqlite3_free (dbMsg);
  }
//------------------------------------------------------------------------------

  void
  Database::execute (const std::string& sql, QueryHandler& sqh)
  {
    command (sql).execute (sqh);
  }
//------------------------------------------------------------------------------

  void
  Database::execute (const std::string& sql, QueryCallback sqc)
  {
    command (sql).execute (sqc);
  }
//------------------------------------------------------------------------------

  Dataset
  Database::select (const std::string& sql)
  {
    return Command (_connection, sql).select ();
  }
//------------------------------------------------------------------------------

  Dataset
  Database::select (const std::string& sql, const Types& types)
  {

    Dataset ds (types);
    execute (sql, ds);
    return ds;

  }
//------------------------------------------------------------------------------

  DbValue
  Database::selectValue (const std::string& sql)
  {

    DbValue retVal (Type::Variant);

    auto cb = [&retVal](QueryRow& queryrows) -> bool
      {
        retVal = queryrows.at (0);
        return false; //exit after first row
      };

    Command cmd (_connection, sql);
    cmd.execute (cb);

    return retVal;
  }
//------------------------------------------------------------------------------

  DbValue
  Database::selectValue (const std::string& sql, Type type)
  {

    DbValue retVal (type);

    auto cb = [&retVal](QueryRow& queryrows) -> bool
      {
        retVal = queryrows.at (0);
        return false; //exit after first row
      };

    Command cmd (_connection, sql);
    cmd.execute (cb);

    return retVal;
  }
//------------------------------------------------------------------------------

  int
  Database::getMostRecentErrCode ()
  {
    return sqlite3_extended_errcode (_connection->sq3ptr);
  }
//------------------------------------------------------------------------------

  std::string
  Database::getMostRecentErrMsg ()
  {
    return std::string (sqlite3_errmsg (_connection->sq3ptr));
  }
//------------------------------------------------------------------------------

  std::string
  Database::getErrStr (int reslutcode)
  {
    return std::string (sqlite3_errstr (reslutcode));
  }
//------------------------------------------------------------------------------

  std::size_t
  Database::getTotalChanges ()
  {
    return sqlite3_total_changes (_connection->sq3ptr);
  }

//------------------------------------------------------------------------------
  std::size_t
  Database::getRecentlyChanged ()
  {
    return sqlite3_changes (_connection->sq3ptr);
  }
//------------------------------------------------------------------------------

  int64_t
  Database::getLastInsertRowid ()
  {
    return sqlite3_last_insert_rowid (_connection->sq3ptr);
  }
//------------------------------------------------------------------------------

  int
  Database::getThreadSafeCompileOption ()
  {
    return sqlite3_threadsafe ();
  }
//------------------------------------------------------------------------------

  auto
  Database::beginTransaction ()
  -> TransactionGuard
  {
    return TransactionGuard { *this };

  }

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
  Database::TransactionGuard::TransactionGuard (Database& db) :
      _db (db),
      _commit (false)
  {
    _db.execute ("BEGIN TRANSACTION");
  }
//------------------------------------------------------------------------------

  Database::TransactionGuard::TransactionGuard (TransactionGuard&& other) :
      _db (other._db),
      _commit (other._commit)
  {
    other._commit = true; // ensure other does not commit in d'tor
  }
//------------------------------------------------------------------------------

  Database::TransactionGuard::~TransactionGuard ()
  {
    if (!_commit)
      _db.execute ("ROLLBACK TRANSACTION");
  }
//------------------------------------------------------------------------------

  void
  Database::TransactionGuard::commit ()
  {
    _db.execute ("COMMIT TRANSACTION");
    _commit = true;
  }

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
}// ns
