/*
 --------------Copyright (c) 2012-2015 H a r a l d  A c h i t z---------------
 -----------< h a r a l d dot a c h i t z at g m a i l dot c o m >------------
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 -----------------------------------------------------------------------------
 */

#include <sq3/dataset.hpp>
#include <sq3/error.hpp>
#include <sqlite3.h>

#include  <stdexcept>
#include  <iterator>
#include  <algorithm>

namespace sq3
{

//------------------------------------------------------------------------------

//------------------------------------------------------------------------------

  Dataset::Dataset () noexcept
  : QueryHandler()
  , _fieldtypes()
  , _names()
    {
    }
//------------------------------------------------------------------------------

  Dataset::Dataset (Types fl) :
      QueryHandler (),
      _fieldtypes (std::move (fl)),
      _names ()
  {
  }
//------------------------------------------------------------------------------
  using namespace std;

  Dataset::Dataset (const Dataset& other) :
      Container<std::vector<DbValues>> (other),
      QueryHandler (),
      _fieldtypes (other._fieldtypes),
      _names (other._names)
  {

  }

//------------------------------------------------------------------------------
  Dataset::Dataset (Dataset&& other) noexcept(
      std::is_nothrow_move_constructible<Container<DbValues>>::value &&
      std::is_nothrow_move_constructible<Types>::value &&
      std::is_nothrow_move_constructible<std::vector<std::string>>::value) :
      Container<std::vector<DbValues>> (std::move (other)),
      QueryHandler (),
      _fieldtypes (move (other._fieldtypes)),
      _names (move (other._names))
  {

  }
//------------------------------------------------------------------------------

  Dataset&
  Dataset::operator= (const Dataset& other)
  {
    _fieldtypes = other._fieldtypes;
    _names = other._names;
    _cont = other._cont;
    return *this;
  }
//------------------------------------------------------------------------------

  Dataset&
  Dataset::operator= (Dataset&& other)
  {
    _fieldtypes = move (other._fieldtypes);
    _names = move (other._names);
    _cont = move (other._cont);
    return *this;
  }
//------------------------------------------------------------------------------

  void
  Dataset::onStart ()
  {
    ; //
  }
//------------------------------------------------------------------------------

  void
  Dataset::onEnd ()
  {
    ;
  }
//------------------------------------------------------------------------------

  bool
  Dataset::onRow (const QueryRow& queryrow)
  {

    if (_names.size () == 0)
      {
        if (_fieldtypes.size () == 0)
          {
            using container_type = Types::container_type;
            Types fieldtypes = container_type (queryrow.size (), Type::Variant);
            _fieldtypes.swap (fieldtypes);
          }
        else if (_fieldtypes.size () != queryrow.size ())
          {
            throw ErrTypeMisMatch (
                "DbValuesTypeList.size != queryrow.getColumnCount()");
          }
        _names = queryrow.getNames ();
      }

    // this will throw if a type does not match.
    _cont.emplace_back (queryrow.getRow (_fieldtypes));

    return true;
  }
//------------------------------------------------------------------------------

  void
  Dataset::reset ()
  {
    _names.clear ();
    _cont.clear ();

  }
//------------------------------------------------------------------------------

  void
  Dataset::reset (const Types& fl)
  {
    _fieldtypes = fl;
    reset ();
  }
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------

  void
  Dataset::merge (const Dataset& other)
  {

    if (not other._names.empty ())
      {
        if (not _names.empty () && _names != other._names)
          throw ErrTypeMisMatch ();
      }

    if (_fieldtypes.size () not_eq other._fieldtypes.size ())
      throw ErrTypeMisMatch ();

    for (std::size_t i = 0; i < _fieldtypes.size (); ++i)
      {
        if (_fieldtypes [i] != Type::Variant)
          {
            if (_fieldtypes [i] != other._fieldtypes [i])
              throw ErrTypeMisMatch ();
          }

      }

    _cont.insert (_cont.end (), other._cont.begin (), other._cont.end ());

  }
//------------------------------------------------------------------------------

  void
  Dataset::merge (const DbValues& other)
  {

    if (_fieldtypes.size () > 0 && _fieldtypes.size () not_eq other.size ())
      {
        throw ErrTypeMisMatch ();
      }

    for (std::size_t i = 0; i < _fieldtypes.size (); ++i)
      {
        if (_fieldtypes [i] != Type::Variant)
          {
            if (_fieldtypes [i] != other [i].getType ())
              {
                throw ErrTypeMisMatch ();
              }
          }
      }

    _cont.push_back (DbValues (other));

  }
//------------------------------------------------------------------------------

  size_t
  Dataset::getIndex (const string& name) const
  {
    using namespace std;
    auto pos = find (_names.begin (), _names.end (), name);
    if (pos == _names.end ())
      throw ErrOutOfRange ("Field name " + name + " not found");

    return distance (_names.begin (), pos);
  }
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
}

