/*
 --------------Copyright (c) 2009-2015 H a r a l d  A c h i t z---------------
 -----------< h a r a l d dot a c h i t z at g m a i l dot c o m >------------
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 -----------------------------------------------------------------------------
 */

#include <sq3/dbvalue.hpp>
#include <sq3/error.hpp>
#include <utility>
#include <ostream>

#ifdef _MSC_VER
#define not !
// yes, I like the not syntax :-) , and I  it is standard in c++11
// until I have a better place I define it her
#endif

namespace sq3
{

  namespace
  {
    template<typename T>
      constexpr
      bool
      oneof (const T&)
      {
        return false;
      }

    template<typename T, typename T1, typename ... REST>
      constexpr
      bool
      oneof (const T& a, const T1& b, REST ... rest)
      {
        return (a == b ? true : oneof (a, rest...));
      }

    struct ensure
    {

      ensure (Type t) :
          type (t)
      {
      } //----------------------------------------------------------------------

      ensure&
      notNull ()
      {
        if (type == Type::Null)
          throw ErrNullValueAccess ();

        return *this;
      } //----------------------------------------------------------------------

      template<typename ... ARGS>
        ensure&
        oneOf (ARGS ... args)
        {
          if (not oneof (type, args...))
            throw ErrTypeMisMatch ();

          return *this;
        } //--------------------------------------------------------------------

      ensure&
      sameAs (Type t)
      {
        if (type != t)
          throw ErrTypeMisMatch ();

        return *this;
      } //----------------------------------------------------------------------

    private:
      Type type;
    };

  } //--------------------------------------------------------------------------
//------------------------------------------------------------------------------

  DbValue::DbValue (Type type) noexcept
  : _type(type == Type::Null ? Type::Variant : type)
  , _storageType(Type::Null)
    {

    }
//------------------------------------------------------------------------------

  DbValue::DbValue (int val, Type type) :
      DbValue (type)
  {
    ensure (type).oneOf (Type::Int, Type::Variant);

    _intstore = val;
    _storageType = Type::Int;
  }
//------------------------------------------------------------------------------

  DbValue::DbValue (int64_t val, Type type) :
      DbValue (type)
  {
    ensure (type).oneOf (Type::Int, Type::Variant);

    _intstore = val;
    _storageType = Type::Int;
  }
//------------------------------------------------------------------------------

  DbValue::DbValue (std::string val, Type type) :
      DbValue (type)
  {
    ensure (type).oneOf (Type::Text, Type::Variant);

    new (&_textstore) std::string (std::move (val));
    _storageType = Type::Text;
  }
//------------------------------------------------------------------------------

  DbValue::DbValue (double val, Type type) :
      DbValue (type)
  {
    ensure (type).oneOf (Type::Real, Type::Variant);

    _realstore = val;
    _storageType = Type::Real;
  }
//------------------------------------------------------------------------------

  DbValue::DbValue (Blob val, Type type) :
      DbValue (type)
  {
    ensure (type).oneOf (Type::Blob, Type::Variant);

    new (&_blobstore) Blob (std::move (val));
    _storageType = Type::Blob;
  }
//------------------------------------------------------------------------------

  DbValue::~DbValue () noexcept
  {
    clearValue ();
  }
//------------------------------------------------------------------------------

  void
  DbValue::clearValue ()
  {
    if (_storageType == Type::Text)
      _textstore.~basic_string<std::string::value_type> ();
    else if (_storageType == Type::Blob)
      _blobstore.~vector<Blob::value_type> ();

    _storageType = Type::Null;

  }
//------------------------------------------------------------------------------

  DbValue::DbValue (const DbValue& other) noexcept
  : _type(other._type)
  , _storageType(other._storageType)
    {

      switch(_storageType)
        {
          case Type::Null:
            break;

          case Type::Int:
            _intstore = other._intstore;
            break;

          case Type::Real:
            _realstore = other._realstore;
            break;

          case Type::Text:
            new(&_textstore) std::string( other._textstore );
            break;

          case Type::Blob:
            new(&_blobstore) Blob( other._blobstore );
            break;

          default: // means variant, terminate (should never happen)
            throw ErrUnexpected("should never reach default in switch");
        }

    }
//------------------------------------------------------------------------------

  DbValue::DbValue (DbValue&& other) noexcept
  : _type(std::move(other._type))
  , _storageType(std::move(other._storageType))
    {
      switch(_storageType)
        {
          case Type::Null:
            break;

          case Type::Int:
            _intstore = std::move(other._intstore);
            break;

          case Type::Real:
            _realstore = std::move(other._realstore);
            break;

          case Type::Text:
            new(&_textstore) std::string( std::move(other._textstore) );
            break;

          case Type::Blob:
            new(&_blobstore) Blob( std::move(other._blobstore) );
            break;

          default: // means variant, terminate (should never happen)
            throw ErrUnexpected("should never reach default in switch");
        }

      // important, set other to null so that clear does not trial to clear
      other._storageType = Type::Null;
    }
//------------------------------------------------------------------------------

  DbValue&
  DbValue::operator= (const DbValue& other)
  {

    if (this->getType () != Type::Variant)
      {
        if (other.getType () == Type::Variant)
          {
            ensure (other._storageType).oneOf (_type, Type::Null);
          }
        else
          {
            ensure (getType ()).sameAs (other.getType ());
          }
      }

    if (_storageType != Type::Null)
      {
        clearValue ();
      }
    // _type = other._type ;
    //NOTE do not change the type,
    // if type is incompatible this line should not be reached!

    _storageType = other._storageType;

    switch (_storageType)
      {
      case Type::Null :
        break;

      case Type::Int :
        _intstore = other._intstore;
        break;

      case Type::Real :
        _realstore = other._realstore;
        break;

      case Type::Text :
        new (&_textstore) std::string (other._textstore);
        break;

      case Type::Blob :
        new (&_blobstore) Blob (other._blobstore);
        break;

      default :
        throw ErrUnexpected ("should never reach default in switch");
      }

    return *this;
  }
//------------------------------------------------------------------------------

  DbValue&
  DbValue::operator= (DbValue&& other)
  {
    if (this->getType () != Type::Variant)
      {
        if (other.getType () == Type::Variant)
          {
            ensure (other._storageType).oneOf (_type, Type::Null);
          }
        else
          {
            ensure (getType ()).sameAs (other.getType ());
          }
      }

    if (_storageType != Type::Null)
      clearValue ();

    // _type = other._type ;
    //NOTE do not change the type,
    // if type is incompatible this line should not be reached!

    _storageType = std::move (other._storageType);

    switch (_storageType)
      {
      case Type::Null :
        break;

      case Type::Int :
        _intstore = std::move (other._intstore);
        break;

      case Type::Real :
        _realstore = std::move (other._realstore);
        break;

      case Type::Text :
        new (&_textstore) std::string (std::move (other._textstore));
        break;

      case Type::Blob :
        new (&_blobstore) Blob (std::move (other._blobstore));
        break;

      default :
        throw ErrUnexpected ("should never reach default in switch");
      }

    // important, set other to null
    other.clearValue ();

    return *this;
  }
//------------------------------------------------------------------------------

  DbValue&
  DbValue::operator= (int val)
  {
    set (val);
    return *this;
  }
//------------------------------------------------------------------------------

  DbValue&
  DbValue::operator= (const int64_t& val)
  {
    set (val);
    return *this;
  }
//------------------------------------------------------------------------------

  DbValue&
  DbValue::operator= (const double& val)
  {
    set (val);
    return *this;
  }
//------------------------------------------------------------------------------

  DbValue&
  DbValue::operator= (const std::string& val)
  {
    set (std::move (val));
    return *this;
  }
//------------------------------------------------------------------------------

  DbValue&
  DbValue::operator= (const Blob& val)
  {
    set (std::move (val));
    return *this;
  }
//------------------------------------------------------------------------------

  void
  DbValue::set (int val)
  {
    // not sure if I leaf this conversion,
    // but its better in text dbval.set(12); it type is real
    if (_type == Type::Real)
      set (static_cast<double> (val));
    else
      set (static_cast<int64_t> (val));
  }
//------------------------------------------------------------------------------

  void
  DbValue::set (int64_t val)
  {
    ensure (_type).oneOf (Type::Int, Type::Variant);

    if (_storageType != Type::Int)
      {
        clearValue ();
        _intstore = val;
        _storageType = Type::Int;
      }
    else
      _intstore = val;
  }
//------------------------------------------------------------------------------

  void
  DbValue::set (double val)
  {
    ensure (_type).oneOf (Type::Real, Type::Variant);

    if (_storageType != Type::Real)
      {
        clearValue ();
        _realstore = val;
        _storageType = Type::Real;
      }
    else
      {
        _realstore = val;
      }
  }
//------------------------------------------------------------------------------

  void
  DbValue::set (const std::string& val)
  {
    ensure (_type).oneOf (Type::Text, Type::Variant);

    if (_storageType != Type::Text)
      {
        clearValue ();
        new (&_textstore) std::string (val);
        _storageType = Type::Text;
      }
    else
      {
        _textstore = val;
      }
  }
//------------------------------------------------------------------------------

  void
  DbValue::set (const Blob& val)
  {
    ensure (_type).oneOf (Type::Blob, Type::Variant);

    if (_storageType != Type::Blob)
      {
        clearValue ();
        new (&_blobstore) Blob (val);
        _storageType = Type::Blob;
      }
    else
      {
        _blobstore = val;
      }
  }
//------------------------------------------------------------------------------

//
  void
  DbValue::setNull ()
  {
    clearValue ();

  }
//------------------------------------------------------------------------------

  const int64_t&
  DbValue::getInt () const
  {
    ensure (_storageType).notNull ().sameAs (Type::Int);

    return _intstore;
  }
//------------------------------------------------------------------------------
  const double&
  DbValue::getReal () const
  {
    ensure (_storageType).notNull ().sameAs (Type::Real);

    return _realstore;
  }
//------------------------------------------------------------------------------

  const std::string&
  DbValue::getText () const
  {
    ensure (_storageType).notNull ().sameAs (Type::Text);

    return _textstore;
  }
//------------------------------------------------------------------------------

  const Blob&
  DbValue::getBlob () const
  {
    ensure (_storageType).notNull ().sameAs (Type::Blob);

    return _blobstore;

  }
//------------------------------------------------------------------------------
  bool
  DbValue::operator== (const int val) const
  {
    if (_storageType == Type::Int)
      return _intstore == val;

    return false;
  }
//------------------------------------------------------------------------------

  bool
  DbValue::operator== (const int64_t& val) const
  {
    if (_storageType == Type::Int)
      return _intstore == val;

    return false;
  }
//------------------------------------------------------------------------------

  bool
  DbValue::operator== (const std::string& val) const
  {
    if (_storageType == Type::Text)
      return _textstore == val;

    return false;
  }
//------------------------------------------------------------------------------

  bool
  DbValue::operator== (const double& val) const
  {
    if (_storageType == Type::Real)
      return _realstore == val;

    return false;
  }
//------------------------------------------------------------------------------

  bool
  DbValue::operator== (const Blob& val) const
  {
    if (_storageType == Type::Blob)
      return _blobstore == val;

    return false;
  }
//------------------------------------------------------------------------------

  bool
  DbValue::operator!= (const int val) const
  {
    return !(*this == val);
  }
//------------------------------------------------------------------------------

  bool
  DbValue::operator!= (const int64_t& val) const
  {
    return !(*this == val);
  }
//------------------------------------------------------------------------------

  bool
  DbValue::operator!= (const std::string& val) const
  {
    return !(*this == val);
  }
//------------------------------------------------------------------------------

  bool
  DbValue::operator!= (const double& val) const
  {
    return !(*this == val);
  }
//------------------------------------------------------------------------------

  bool
  DbValue::operator!= (const Blob& val) const
  {
    return !(*this == val);
  }
//------------------------------------------------------------------------------

  std::string
  DbValue::ejectText ()
  {
    ensure (_storageType).notNull ().sameAs (Type::Text);

    auto tmp = std::move (_textstore);
    setNull ();
    return tmp;
  }
//------------------------------------------------------------------------------

  Blob
  DbValue::ejectBlob ()
  {
    ensure (_storageType).notNull ().sameAs (Type::Blob);

    auto tmp = std::move (_blobstore);
    setNull ();
    return tmp;
  }
//------------------------------------------------------------------------------

  DbValues&
  DbValues::operator= (const DbValues& row)
  {
    if (row.size () != size ())
      throw ErrTypeMisMatch ();

    _cont = row._cont;
    return *this;
  }
//------------------------------------------------------------------------------

  DbValues&
  DbValues::operator= (DbValues&& row)
  {
    if (row.size () != size ())
      throw ErrTypeMisMatch ();

    _cont = std::move (row._cont);
    return *this;
  }
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
  std::ostream&
  operator<< (std::ostream& stm, const sq3::DbValue& v)
  {

    switch (v.getStorageType ())
      {
      case Type::Null :
        stm << "<NULL>";
        break;

      case Type::Int :
        stm << v.getInt ();
        break;

      case Type::Real :
        stm << v.getReal ();
        break;

      case Type::Text :
        stm << v.getText ();
        break;

      case Type::Blob :
        stm << "<BLOB>";
        break;
      default :
        stm << "unknown storage type !!";
        break;
      }

    return stm;
  }
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
}// ns

