/*
 --------------Copyright (c) 2009-2015 H a r a l d  A c h i t z---------------
 -----------< h a r a l d dot a c h i t z at g m a i l dot c o m >------------
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 -----------------------------------------------------------------------------
 */

#include <sq3/error.hpp>
#include <sq3/queryhandler.hpp>
#include <sqlite3.h>

namespace sq3
{

  namespace
  {

    // std::string
    // getString(sqlite3_stmt* stmt, int idx, std::size_t* size = nullptr)
    // {
    //   std::string retval;
    //   const char* first = (const char*) sqlite3_column_text(stmt, idx);
    //   std::size_t s = sqlite3_column_bytes(stmt, idx);
    //   if( s > 0)
    //     retval.assign(first, s);
    //
    //   if( size != nullptr )
    //     *size = s;
    //   return retval;
    // }//---------------------------------------------------------------------

    std::string
    getString (sqlite3_stmt* stmt, int idx)
    {
      const char* first = (const char*) sqlite3_column_text (stmt, idx);
      std::size_t s = sqlite3_column_bytes (stmt, idx);
      return s > 0 ? std::string (first, s) : std::string ();
    } //-----------------------------------------------------------------------

//  std::size_t
//  intoString(sqlite3_stmt* stmt, int idx, std::string& val)
//  {
//    const char* first = (const char*) sqlite3_column_text(stmt, idx);
//    std::size_t size = sqlite3_column_bytes(stmt, idx);
//    if( size )
//      val = first;
//    else
//      val = std::string();
//    return size;
//  }//------------------------------------------------------------------------

//  int
//  getInt(sqlite3_stmt* stmt, int idx, std::size_t* size)
//  {
//    int retVal = sqlite3_column_int(stmt, idx);
//    *size = sqlite3_column_bytes(stmt, idx);
//    return retVal;
//  }//------------------------------------------------------------------------

//  int
//  getInt(sqlite3_stmt* stmt, int idx)
//  {
//    return sqlite3_column_int(stmt, idx);
//  }//------------------------------------------------------------------------

//  std::size_t
//  intoInt(sqlite3_stmt* stmt, int idx, int& val)
//  {
//    val = sqlite3_column_int(stmt, idx);
//    return sqlite3_column_bytes(stmt, idx);
//  }//------------------------------------------------------------------------

    // int64_t
    // getInt64(sqlite3_stmt* stmt, int idx, std::size_t* size)
    // {
    //   int64_t retVal = sqlite3_column_int64(stmt, idx);
    //   *size = sqlite3_column_bytes(stmt, idx);
    //   return retVal;
    // }//---------------------------------------------------------------------

    int64_t
    getInt64 (sqlite3_stmt* stmt, int idx)
    {
      return sqlite3_column_int64 (stmt, idx);
    } //-----------------------------------------------------------------------

//  std::size_t
//  intoInt64(sqlite3_stmt* stmt, int idx, int64_t& val)
//  {
//    val = sqlite3_column_int64(stmt, idx);
//    return sqlite3_column_bytes(stmt, idx);
//  }//------------------------------------------------------------------------

    // double
    // getReal(sqlite3_stmt* stmt, int idx, std::size_t* size)
    // {
    //   double retVal = sqlite3_column_double(stmt, idx);
    //   *size = sqlite3_column_bytes(stmt, idx);
    //   return retVal;
    // }//---------------------------------------------------------------------

    double
    getReal (sqlite3_stmt* stmt, int idx)
    {
      return sqlite3_column_double (stmt, idx);
    } //-----------------------------------------------------------------------

//  std::size_t
//  intoReal(sqlite3_stmt* stmt, int idx, double& val)
//  {
//    val = sqlite3_column_double(stmt, idx);
//    return sqlite3_column_bytes(stmt, idx);
//  }//------------------------------------------------------------------------

    // Blob
    // getBlob(sqlite3_stmt* stmt, int idx, std::size_t* size)
    // {
    //   using value_type = Blob::value_type ;
    //   const value_type* first =
    //       static_cast<const value_type*>(sqlite3_column_blob(stmt, idx));
    //   std::size_t s = sqlite3_column_bytes(stmt, idx);
    //   *size = s;
    //   return s > 0 ? Blob(first, first + s): Blob();
    // }//---------------------------------------------------------------------

    Blob
    getBlob (sqlite3_stmt* stmt, int idx)
    {
      using value_type = Blob::value_type ;
      const value_type* first =
          static_cast<const value_type*> (sqlite3_column_blob (stmt, idx));
      std::size_t s = sqlite3_column_bytes (stmt, idx);
      return s > 0 ? Blob (first, first + s) : Blob ();
    } //-----------------------------------------------------------------------

//  std::size_t
//  intoBlob(sqlite3_stmt* stmt, int idx, Blob& val)
//  {
//    const Blob::value_type* first =
//        static_cast<const Blob::value_type*>(sqlite3_column_blob(stmt, idx))
//
//    std::size_t size = sqlite3_column_bytes(stmt, idx);
//    if( size )
//      {
//        val.resize(size);
//        val.assign(first, first + size);
//      }
//    else
//      {
//        val.clear();
//      }
//    return size;
//
//  }//------------------------------------------------------------------------

    inline DbValue
    getVal (sqlite3_stmt* stmt, int idx, Type type = Type::Variant)
    {
      switch (sqlite3_column_type (stmt, idx))
        {
        case SQLITE_INTEGER :
          return DbValue (getInt64 (stmt, idx), type);
          break;
        case SQLITE_FLOAT :
          return DbValue (getReal (stmt, idx), type);
          break;
        case SQLITE_TEXT :
          return DbValue (getString (stmt, idx), type);
          break;
        case SQLITE_BLOB :
          return DbValue (getBlob (stmt, idx), type);
          break;
        case SQLITE_NULL :
          return DbValue (type);
          break;
        default :
#ifdef DEBUG
          throw ErrUnexpected ("should never reach this line");
#endif
          return DbValue (type);
          break;
        }
      return DbValue (type); //for the lexer
    } //-----------------------------------------------------------------------

  } // ns----------------------------------------------------------------------

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------

  int
  QueryRow::getColumnCount () const
  {
    return sqlite3_column_count (_stmt);
  }
//-----------------------------------------------------------------------------

  std::string
  QueryRow::getColumnName (int idx) const
  {
    const char* name = sqlite3_column_name (_stmt, idx);
    if (name)
      return std::string (name);
    else
      return std::string ();
  }
//-----------------------------------------------------------------------------

  DbValue
  QueryRow::at (int idx) const
  {
    if (idx < 0 || !(idx < getColumnCount ()))
      throw ErrOutOfRange ("column index out of range");

    return getVal (_stmt, idx);

  }
//-----------------------------------------------------------------------------

  DbValue
  QueryRow::operator[] (int idx) const
  {
    return getVal (_stmt, idx);
  }
//-----------------------------------------------------------------------------

  std::vector<std::string>
  QueryRow::getNames () const
  {
    std::vector<std::string> names (size ());
    for (int i = 0; i < getColumnCount (); ++i)
      {
        names [i] = getColumnName (i);
      }

    return names;
  }
//-----------------------------------------------------------------------------

  sqlite3_stmt*
  QueryRow::get_stmt () const
  {
    return _stmt;
  }
//-----------------------------------------------------------------------------

  DbValues
  QueryRow::getRow () const
  {
    DbValues::container_type v;
    for (int i = 0; i < getColumnCount (); ++i)
      {
        v.push_back (getVal (_stmt, i));
      }
    return DbValues (std::move (v));

  }
//-----------------------------------------------------------------------------

/// return all columns at once
  DbValues
  QueryRow::getRow (const Types& types) const
  {
    if (types.size () != size ())
      {
        throw ErrTypeMisMatch ("types.size () != getColumnCount ()");
      }

    DbValues::container_type v;
    for (int i = 0; i < getColumnCount (); ++i)
      {
        v.push_back (getVal (_stmt, i, types [i]));
      }
    return DbValues (std::move (v));

  }
//-----------------------------------------------------------------------------

  bool
  QueryHandler::handleRow (sqlite3_stmt* stmt)
  {
    return onRow (QueryRow (stmt));
  }
//-----------------------------------------------------------------------------

}// ns
