#include "../../include/sq3/queryhandler.hpp"

#include <a4testing.hpp>

#include <sq3/database.hpp>
#include <sq3/error.hpp>

namespace sq3 {
namespace {

Database
testdb ()
{

  Database db { ":memory:" };
  db.execute ("CREATE TABLE tbl(f1 INTEGER, f2 TEXT);"
              "INSERT INTO tbl (f1, f2 ) VALUES (1,'one');"
              "INSERT INTO tbl (f1, f2 ) VALUES (2,'two');"
              "INSERT INTO tbl (f1, f2 ) VALUES (3,'three');");

  return std::move (db);
}
//------------------------------------------------------------------------------

void
checkRowReturnValHandler ()
{

  auto db = testdb ();

  struct : public QueryHandler
  {
    int rowcount = 0;
    bool
    onRow (const QueryRow&)
    {
      rowcount += 1;
      return true;
    }
  } handleTrue;

  BOOST_REQUIRE_EQUAL(handleTrue.rowcount, 0);
  db.command ("SELECT * FROM tbl;").execute (handleTrue);
  BOOST_REQUIRE_EQUAL(handleTrue.rowcount, 3); // OnHandleRow called 3 times

  // return false after first call,
  struct : public QueryHandler
  {
    int rowcount = 0;
    bool
    onRow (const QueryRow&)
    {
      rowcount += 1;
      return false;
    }
  } handleFalse;

  BOOST_REQUIRE_EQUAL(handleFalse.rowcount, 0);
  db.command ("SELECT * FROM tbl;").execute (handleFalse);
  BOOST_REQUIRE_EQUAL(handleFalse.rowcount, 1); // OnHandleRow  called once

}
//------------------------------------------------------------------------------

void
checkRowReturnValCallback ()
{
  auto db = testdb ();

  int countTrue = 0;
  auto handleTrue = [&](QueryRow& ) -> bool
    {
      countTrue+=1;
      return true;
    };

  int countFalse = 0;
  auto handleFalse = [&](QueryRow& ) -> bool
    {
      countFalse+=1;
      return false;
    };

  auto cmd = db.command ("SELECT * FROM tbl;");
  cmd.execute (handleTrue);
  cmd.execute (handleFalse);

  BOOST_REQUIRE_EQUAL(countTrue, 3);
  BOOST_REQUIRE_EQUAL(countFalse, 1);

}
//------------------------------------------------------------------------------
void
checkColumnNamesAndCount ()
{

  auto testcb = [&](QueryRow& queryrow) -> bool
    {
      BOOST_CHECK( queryrow.getColumnCount() == 2 );
      BOOST_CHECK( queryrow.getColumnName(0) == "f1" );
      BOOST_CHECK( queryrow.getColumnName(1) == "f2" );
      BOOST_CHECK( queryrow.getColumnName(2) == "" );

      return false;
    };
  auto db = testdb ();
  db.execute ("SELECT * FROM tbl;", testcb);
}
//------------------------------------------------------------------------------

void
checkColumnAt ()
{
  auto testcb = [&](QueryRow& queryrow) -> bool
    {
      auto names = queryrow.getNames();

      BOOST_CHECK_NO_THROW( queryrow.at(0) );
      BOOST_CHECK( names.at(0) == "f1" );

      BOOST_CHECK_NO_THROW( queryrow.at(1) );
      BOOST_CHECK( names.at(1) == "f2" );

      BOOST_CHECK_THROW( queryrow.at(2) , ErrOutOfRange );

      return false;
    };
  auto db = testdb ();
  db.execute ("SELECT * FROM tbl;", testcb);
}
//------------------------------------------------------------------------------

void
checkGetRow ()
{
  auto testcb = [&](QueryRow& queryrow) -> bool
    {
      BOOST_CHECK_THROW( queryrow.at(2) , ErrOutOfRange );
      
      BOOST_REQUIRE_NO_THROW ( queryrow.getRow ());
      BOOST_REQUIRE ( queryrow.getRow ().size () == 2);

      Types wrongTypes1 { {Type::Real } };
      Types wrongTypes2 { {Type::Real, Type::Blob, Type::Blob} };
      Types wrongTypes3 { {Type::Real, Type::Blob} };
      Types rightTypes { {Type::Int, Type::Text} };

      BOOST_REQUIRE_THROW( queryrow.getRow (wrongTypes1), ErrTypeMisMatch );
      BOOST_REQUIRE_THROW( queryrow.getRow (wrongTypes2), ErrTypeMisMatch );
      BOOST_REQUIRE_THROW( queryrow.getRow (wrongTypes3), ErrTypeMisMatch );
      BOOST_REQUIRE_NO_THROW( queryrow.getRow (rightTypes));
      
      auto variants = queryrow.getRow ();
      BOOST_REQUIRE ( variants [0].getType () == Type::Variant);
      BOOST_REQUIRE ( variants [1].getType () == Type::Variant);

      auto typed = queryrow.getRow (rightTypes);
      BOOST_REQUIRE ( typed [0].getType () == Type::Int);
      BOOST_REQUIRE ( typed [1].getType () == Type::Text);
      return false;
    };
  auto db = testdb ();
  db.execute ("SELECT * FROM tbl;", testcb);
}
//------------------------------------------------------------------------------
void 
forCoverage()
{

  auto db = testdb ();
  db.execute ("SELECT * FROM tbl;", 
	      [&](QueryRow& qr)
	      {
		BOOST_CHECK( qr.get_stmt () != nullptr );
		return false;
	      } );

}

//------------------------------------------------------------------------------

a4TestAdd(
    a4test::suite("QueryHandler")
    .addTest("return handler", checkRowReturnValHandler)
    .addTest("value callback", checkRowReturnValCallback)
    .addTest("column names & count", checkColumnNamesAndCount)
    .addTest("column index", checkColumnAt)
    .addTest("getRow", checkGetRow)
    .addTest("coverage", forCoverage)
);

}
}

